@@ -1,44 +1,41 @@
 use crate::{
-    compilation::{compile_validators, context::CompilationContext},
-    error::{error, no_error, ErrorIterator, ValidationError},
+    compiler,
+    error::ValidationError,
     keywords::CompilationResult,
+    node::SchemaNode,
     output::BasicOutput,
-    paths::{InstancePath, JSONPointer},
-    primitive_type::PrimitiveType,
-    schema_node::SchemaNode,
-    validator::{format_iter_of_validators, PartialApplication, Validate},
+    paths::{LazyLocation, Location},
+    types::JsonType,
+    validator::{PartialApplication, Validate},
 };
 use serde_json::{Map, Value};
 
 pub(crate) struct OneOfValidator {
     schemas: Vec<SchemaNode>,
-    schema_path: JSONPointer,
+    location: Location,
 }
 
 impl OneOfValidator {
     #[inline]
-    pub(crate) fn compile<'a>(
-        schema: &'a Value,
-        context: &CompilationContext,
-    ) -> CompilationResult<'a> {
+    pub(crate) fn compile<'a>(ctx: &compiler::Context, schema: &'a Value) -> CompilationResult<'a> {
         if let Value::Array(items) = schema {
-            let keyword_context = context.with_path("oneOf");
+            let ctx = ctx.new_at_location("oneOf");
             let mut schemas = Vec::with_capacity(items.len());
             for (idx, item) in items.iter().enumerate() {
-                let item_context = keyword_context.with_path(idx);
-                let node = compile_validators(item, &item_context)?;
-                schemas.push(node)
+                let ctx = ctx.new_at_location(idx);
+                let node = compiler::compile(&ctx, ctx.as_resource_ref(item))?;
+                schemas.push(node);
             }
             Ok(Box::new(OneOfValidator {
                 schemas,
-                schema_path: keyword_context.into_pointer(),
+                location: ctx.location().clone(),
             }))
         } else {
             Err(ValidationError::single_type_error(
-                JSONPointer::default(),
-                context.clone().into_pointer(),
+                Location::new(),
+                ctx.location().clone(),
                 schema,
-                PrimitiveType::Array,
+                JsonType::Array,
             ))
         }
     }
@@ -54,7 +51,7 @@ impl OneOfValidator {
         first_valid_idx
     }
 
-    #[allow(clippy::integer_arithmetic)]
+    #[allow(clippy::arithmetic_side_effects)]
     fn are_others_valid(&self, instance: &Value, idx: usize) -> bool {
         // `idx + 1` will not overflow, because the maximum possible value there is `usize::MAX - 1`
         // For example we have `usize::MAX` schemas and only the last one is valid, then
@@ -69,43 +66,47 @@ impl OneOfValidator {
 impl Validate for OneOfValidator {
     fn is_valid(&self, instance: &Value) -> bool {
         let first_valid_idx = self.get_first_valid(instance);
-        first_valid_idx.map_or(false, |idx| !self.are_others_valid(instance, idx))
+        first_valid_idx.is_some_and(|idx| !self.are_others_valid(instance, idx))
     }
-    fn validate<'instance>(
+    fn validate<'i>(
         &self,
-        instance: &'instance Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'instance> {
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
         let first_valid_idx = self.get_first_valid(instance);
         if let Some(idx) = first_valid_idx {
             if self.are_others_valid(instance, idx) {
-                return error(ValidationError::one_of_multiple_valid(
-                    self.schema_path.clone(),
-                    instance_path.into(),
+                return Err(ValidationError::one_of_multiple_valid(
+                    self.location.clone(),
+                    location.into(),
                     instance,
+                    self.schemas
+                        .iter()
+                        .map(|schema| schema.iter_errors(instance, location).collect())
+                        .collect(),
                 ));
             }
-            no_error()
+            Ok(())
         } else {
-            error(ValidationError::one_of_not_valid(
-                self.schema_path.clone(),
-                instance_path.into(),
+            Err(ValidationError::one_of_not_valid(
+                self.location.clone(),
+                location.into(),
                 instance,
+                self.schemas
+                    .iter()
+                    .map(|schema| schema.iter_errors(instance, location).collect())
+                    .collect(),
             ))
         }
     }
-    fn apply<'a>(
-        &'a self,
-        instance: &Value,
-        instance_path: &InstancePath,
-    ) -> PartialApplication<'a> {
+    fn apply<'a>(&'a self, instance: &Value, location: &LazyLocation) -> PartialApplication<'a> {
         let mut failures = Vec::new();
         let mut successes = Vec::new();
         for node in &self.schemas {
-            match node.apply_rooted(instance, instance_path) {
+            match node.apply_rooted(instance, location) {
                 output @ BasicOutput::Valid(..) => successes.push(output),
                 output @ BasicOutput::Invalid(..) => failures.push(output),
-            };
+            }
         }
         if successes.len() == 1 {
             let success = successes.remove(0);
@@ -120,23 +121,13 @@ impl Validate for OneOfValidator {
     }
 }
 
-impl core::fmt::Display for OneOfValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(
-            f,
-            "oneOf: [{}]",
-            format_iter_of_validators(self.schemas.iter().map(SchemaNode::validators))
-        )
-    }
-}
-
 #[inline]
 pub(crate) fn compile<'a>(
+    ctx: &compiler::Context,
     _: &'a Map<String, Value>,
     schema: &'a Value,
-    context: &CompilationContext,
 ) -> Option<CompilationResult<'a>> {
-    Some(OneOfValidator::compile(schema, context))
+    Some(OneOfValidator::compile(ctx, schema))
 }
 
 #[cfg(test)]
@@ -147,7 +138,7 @@ mod tests {
 
     #[test_case(&json!({"oneOf": [{"type": "string"}]}), &json!(0), "/oneOf")]
     #[test_case(&json!({"oneOf": [{"type": "string"}, {"maxLength": 3}]}), &json!(""), "/oneOf")]
-    fn schema_path(schema: &Value, instance: &Value, expected: &str) {
-        tests_util::assert_schema_path(schema, instance, expected)
+    fn location(schema: &Value, instance: &Value, expected: &str) {
+        tests_util::assert_schema_location(schema, instance, expected);
     }
 }
