@@ -1,44 +1,41 @@
 use crate::{
-    compilation::{context::CompilationContext, JSONSchema},
-    error::{error, no_error, ErrorIterator, ValidationError},
+    compiler,
+    error::{no_error, ErrorIterator, ValidationError},
     keywords::CompilationResult,
-    paths::{InstancePath, JSONPointer},
-    primitive_type::PrimitiveType,
+    paths::{LazyLocation, Location},
+    types::JsonType,
     validator::Validate,
 };
 use serde_json::{Map, Value};
 
 pub(crate) struct RequiredValidator {
     required: Vec<String>,
-    schema_path: JSONPointer,
+    location: Location,
 }
 
 impl RequiredValidator {
     #[inline]
-    pub(crate) fn compile(items: &[Value], schema_path: JSONPointer) -> CompilationResult {
+    pub(crate) fn compile(items: &[Value], location: Location) -> CompilationResult {
         let mut required = Vec::with_capacity(items.len());
         for item in items {
             match item {
                 Value::String(string) => required.push(string.clone()),
                 _ => {
                     return Err(ValidationError::single_type_error(
-                        JSONPointer::default(),
-                        schema_path,
+                        Location::new(),
+                        location,
                         item,
-                        PrimitiveType::String,
+                        JsonType::String,
                     ))
                 }
             }
         }
-        Ok(Box::new(RequiredValidator {
-            required,
-            schema_path,
-        }))
+        Ok(Box::new(RequiredValidator { required, location }))
     }
 }
 
 impl Validate for RequiredValidator {
-    fn is_valid(&self, _: &JSONSchema, instance: &Value) -> bool {
+    fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Object(item) = instance {
             self.required
                 .iter()
@@ -48,19 +45,34 @@ impl Validate for RequiredValidator {
         }
     }
 
-    fn validate<'a, 'b>(
+    fn validate<'i>(
         &self,
-        _: &'a JSONSchema,
-        instance: &'b Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'b> {
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
+        if let Value::Object(item) = instance {
+            for property_name in &self.required {
+                if !item.contains_key(property_name) {
+                    return Err(ValidationError::required(
+                        self.location.clone(),
+                        location.into(),
+                        instance,
+                        // Value enum is needed for proper string escaping
+                        Value::String(property_name.clone()),
+                    ));
+                }
+            }
+        }
+        Ok(())
+    }
+    fn iter_errors<'i>(&self, instance: &'i Value, location: &LazyLocation) -> ErrorIterator<'i> {
         if let Value::Object(item) = instance {
             let mut errors = vec![];
             for property_name in &self.required {
                 if !item.contains_key(property_name) {
                     errors.push(ValidationError::required(
-                        self.schema_path.clone(),
-                        instance_path.into(),
+                        self.location.clone(),
+                        location.into(),
                         instance,
                         // Value enum is needed for proper string escaping
                         Value::String(property_name.clone()),
@@ -75,47 +87,40 @@ impl Validate for RequiredValidator {
     }
 }
 
-impl core::fmt::Display for RequiredValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "required: [{}]", self.required.join(", "))
-    }
-}
-
 pub(crate) struct SingleItemRequiredValidator {
     value: String,
-    schema_path: JSONPointer,
+    location: Location,
 }
 
 impl SingleItemRequiredValidator {
     #[inline]
-    pub(crate) fn compile(value: &str, schema_path: JSONPointer) -> CompilationResult {
+    pub(crate) fn compile(value: &str, location: Location) -> CompilationResult {
         Ok(Box::new(SingleItemRequiredValidator {
             value: value.to_string(),
-            schema_path,
+            location,
         }))
     }
 }
 
 impl Validate for SingleItemRequiredValidator {
-    fn validate<'a, 'b>(
+    fn validate<'i>(
         &self,
-        schema: &'a JSONSchema,
-        instance: &'b Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'b> {
-        if !self.is_valid(schema, instance) {
-            return error(ValidationError::required(
-                self.schema_path.clone(),
-                instance_path.into(),
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
+        if !self.is_valid(instance) {
+            return Err(ValidationError::required(
+                self.location.clone(),
+                location.into(),
                 instance,
                 // Value enum is needed for proper string escaping
                 Value::String(self.value.clone()),
             ));
         }
-        no_error()
+        Ok(())
     }
 
-    fn is_valid(&self, _: &JSONSchema, instance: &Value) -> bool {
+    fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Object(item) = instance {
             item.contains_key(&self.value)
         } else {
@@ -124,51 +129,42 @@ impl Validate for SingleItemRequiredValidator {
     }
 }
 
-impl core::fmt::Display for SingleItemRequiredValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "required: [{}]", self.value)
-    }
-}
-
 #[inline]
 pub(crate) fn compile<'a>(
+    ctx: &compiler::Context,
     _: &'a Map<String, Value>,
     schema: &'a Value,
-    context: &CompilationContext,
 ) -> Option<CompilationResult<'a>> {
-    let schema_path = context.as_pointer_with("required");
-    compile_with_path(schema, schema_path)
+    let location = ctx.location().join("required");
+    compile_with_path(schema, location)
 }
 
 #[inline]
-pub(crate) fn compile_with_path(
-    schema: &Value,
-    schema_path: JSONPointer,
-) -> Option<CompilationResult> {
+pub(crate) fn compile_with_path(schema: &Value, location: Location) -> Option<CompilationResult> {
     // IMPORTANT: If this function will ever return `None`, adjust `dependencies.rs` accordingly
     match schema {
         Value::Array(items) => {
             if items.len() == 1 {
                 let item = &items[0];
                 if let Value::String(item) = item {
-                    Some(SingleItemRequiredValidator::compile(item, schema_path))
+                    Some(SingleItemRequiredValidator::compile(item, location))
                 } else {
                     Some(Err(ValidationError::single_type_error(
-                        JSONPointer::default(),
-                        schema_path,
+                        Location::new(),
+                        location,
                         item,
-                        PrimitiveType::String,
+                        JsonType::String,
                     )))
                 }
             } else {
-                Some(RequiredValidator::compile(items, schema_path))
+                Some(RequiredValidator::compile(items, location))
             }
         }
         _ => Some(Err(ValidationError::single_type_error(
-            JSONPointer::default(),
-            schema_path,
+            Location::new(),
+            location,
             schema,
-            PrimitiveType::Array,
+            JsonType::Array,
         ))),
     }
 }
@@ -181,7 +177,7 @@ mod tests {
 
     #[test_case(&json!({"required": ["a"]}), &json!({}), "/required")]
     #[test_case(&json!({"required": ["a", "b"]}), &json!({}), "/required")]
-    fn schema_path(schema: &Value, instance: &Value, expected: &str) {
-        tests_util::assert_schema_path(schema, instance, expected)
+    fn location(schema: &Value, instance: &Value, expected: &str) {
+        tests_util::assert_schema_location(schema, instance, expected);
     }
 }
