@@ -1,9 +1,7 @@
 //! Implementation of json schema output formats specified in <https://json-schema.org/draft/2020-12/json-schema-core.html#rfc.section.12.2>
 //!
-//! Currently the "flag" and "basic" formats are supported. The "flag" format is
-//! idential to the [`JSONSchema::is_valid`] method and so is uninteresting. The
-//! main contribution of this module is [`Output::basic`]. See the documentation
-//! of that method for more information.
+//! Currently the "basic" formats is supported. The main contribution of this module is [`Output::basic`].
+//! See the documentation of that method for more information.
 
 use std::{
     borrow::Cow,
@@ -13,31 +11,29 @@ use std::{
     ops::AddAssign,
 };
 
-use crate::{validator::PartialApplication, ValidationError};
+use crate::{paths::Location, validator::PartialApplication, ValidationError};
 use ahash::AHashMap;
+use referencing::Uri;
 use serde::ser::SerializeMap;
 
-use crate::{
-    paths::{AbsolutePath, InstancePath, JSONPointer},
-    schema_node::SchemaNode,
-    JSONSchema,
-};
+use crate::{node::SchemaNode, paths::LazyLocation, Validator};
 
-/// The output format resulting from the application of a schema. This can be
-/// converted into various representations based on the definitions in
+/// The output format resulting from the application of a schema.
+///
+/// This can be converted into various representations based on the definitions in
 /// <https://json-schema.org/draft/2020-12/json-schema-core.html#rfc.section.12.2>
 ///
 /// Currently only the "flag" and "basic" output formats are supported
 #[derive(Debug, Clone)]
 pub struct Output<'a, 'b> {
-    schema: &'a JSONSchema,
+    schema: &'a Validator,
     root_node: &'a SchemaNode,
     instance: &'b serde_json::Value,
 }
 
-impl<'a, 'b> Output<'a, 'b> {
+impl<'a> Output<'a, '_> {
     pub(crate) const fn new<'c, 'd>(
-        schema: &'c JSONSchema,
+        schema: &'c Validator,
         root_node: &'c SchemaNode,
         instance: &'d serde_json::Value,
     ) -> Output<'c, 'd> {
@@ -69,20 +65,21 @@ impl<'a, 'b> Output<'a, 'b> {
     /// and where in the instance). The difference between the
     /// `BasicOutput::Valid` and `BasicOutput::Invalid` cases is the value which
     /// is associated with each `OutputUnit`. For `Valid` outputs the value is
-    /// an annotation, whilst for `Invalid` outputs it's an `ErrorDescription`
+    /// an annotation, whilst for `Invalid` outputs it's an [`ErrorDescription`]
     /// (a `String` really).
     ///
     /// # Examples
     ///
     /// ```rust
-    /// # use crate::jsonschema::{Draft, output::{Output, BasicOutput}, JSONSchema};
-    /// # let schema_json = serde_json::json!({
+    /// # use jsonschema::BasicOutput;
+    /// # use serde_json::json;
+    /// # let schema = json!({
     /// #     "title": "string value",
     /// #     "type": "string"
     /// # });
-    /// # let instance = serde_json::json!{"some string"};
-    /// # let schema = JSONSchema::options().compile(&schema_json).unwrap();
-    /// let output: BasicOutput = schema.apply(&instance).basic();
+    /// # let instance = json!("some string");
+    /// # let validator = jsonschema::validator_for(&schema).expect("Invalid schema");
+    /// let output = validator.apply(&instance).basic();
     /// match output {
     ///     BasicOutput::Valid(annotations) => {
     ///         for annotation in annotations {
@@ -107,7 +104,7 @@ impl<'a, 'b> Output<'a, 'b> {
     #[must_use]
     pub fn basic(&self) -> BasicOutput<'a> {
         self.root_node
-            .apply_rooted(self.instance, &InstancePath::new())
+            .apply_rooted(self.instance, &LazyLocation::new())
     }
 }
 
@@ -121,7 +118,7 @@ pub enum BasicOutput<'a> {
     Invalid(VecDeque<OutputUnit<ErrorDescription>>),
 }
 
-impl<'a> BasicOutput<'a> {
+impl BasicOutput<'_> {
     /// A shortcut to check whether the output represents passed validation.
     #[must_use]
     pub const fn is_valid(&self) -> bool {
@@ -140,24 +137,24 @@ impl<'a> From<OutputUnit<Annotations<'a>>> for BasicOutput<'a> {
     }
 }
 
-impl<'a> AddAssign for BasicOutput<'a> {
+impl AddAssign for BasicOutput<'_> {
     fn add_assign(&mut self, rhs: Self) {
         match (&mut *self, rhs) {
             (BasicOutput::Valid(ref mut anns), BasicOutput::Valid(anns_rhs)) => {
                 anns.extend(anns_rhs);
             }
             (BasicOutput::Valid(..), BasicOutput::Invalid(errors)) => {
-                *self = BasicOutput::Invalid(errors)
+                *self = BasicOutput::Invalid(errors);
             }
             (BasicOutput::Invalid(..), BasicOutput::Valid(..)) => {}
             (BasicOutput::Invalid(errors), BasicOutput::Invalid(errors_rhs)) => {
-                errors.extend(errors_rhs)
+                errors.extend(errors_rhs);
             }
         }
     }
 }
 
-impl<'a> Sum for BasicOutput<'a> {
+impl Sum for BasicOutput<'_> {
     fn sum<I: Iterator<Item = Self>>(iter: I) -> Self {
         let result = BasicOutput::Valid(VecDeque::new());
         iter.fold(result, |mut acc, elem| {
@@ -167,7 +164,7 @@ impl<'a> Sum for BasicOutput<'a> {
     }
 }
 
-impl<'a> Default for BasicOutput<'a> {
+impl Default for BasicOutput<'_> {
     fn default() -> Self {
         BasicOutput::Valid(VecDeque::new())
     }
@@ -194,24 +191,24 @@ impl<'a> FromIterator<BasicOutput<'a>> for PartialApplication<'a> {
     }
 }
 
-/// An output unit is a reference to a place in a schema and a place in an
-/// instance along with some value associated to that place. For annotations the
-/// value will be an [`Annotations`] and for errors it will be an
+/// A reference to a place in a schema and a place in an instance along with some value associated to that place.
+///
+/// For annotations the value will be an [`Annotations`] and for errors it will be an
 /// [`ErrorDescription`]. See the documentation for [`Output::basic`] for a
 /// detailed example.
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct OutputUnit<T> {
-    keyword_location: JSONPointer,
-    instance_location: JSONPointer,
-    absolute_keyword_location: Option<AbsolutePath>,
+    keyword_location: Location,
+    instance_location: Location,
+    absolute_keyword_location: Option<Uri<String>>,
     value: T,
 }
 
 impl<T> OutputUnit<T> {
     pub(crate) const fn annotations(
-        keyword_location: JSONPointer,
-        instance_location: JSONPointer,
-        absolute_keyword_location: Option<AbsolutePath>,
+        keyword_location: Location,
+        instance_location: Location,
+        absolute_keyword_location: Option<Uri<String>>,
         annotations: Annotations<'_>,
     ) -> OutputUnit<Annotations<'_>> {
         OutputUnit {
@@ -223,9 +220,9 @@ impl<T> OutputUnit<T> {
     }
 
     pub(crate) const fn error(
-        keyword_location: JSONPointer,
-        instance_location: JSONPointer,
-        absolute_keyword_location: Option<AbsolutePath>,
+        keyword_location: Location,
+        instance_location: Location,
+        absolute_keyword_location: Option<Uri<String>>,
         error: ErrorDescription,
     ) -> OutputUnit<ErrorDescription> {
         OutputUnit {
@@ -236,19 +233,21 @@ impl<T> OutputUnit<T> {
         }
     }
 
-    ///  The location in the schema of the keyword
-    pub const fn keyword_location(&self) -> &JSONPointer {
+    /// The location in the schema of the keyword
+    pub const fn keyword_location(&self) -> &Location {
         &self.keyword_location
     }
 
-    ///  The absolute location in the schema of the keyword. This will be
-    ///  different to `keyword_location` if the schema is a resolved reference.
-    pub const fn absolute_keyword_location(&self) -> &Option<AbsolutePath> {
-        &self.absolute_keyword_location
+    /// The absolute location in the schema of the keyword. This will be
+    /// different to `keyword_location` if the schema is a resolved reference.
+    pub fn absolute_keyword_location(&self) -> Option<Uri<&str>> {
+        self.absolute_keyword_location
+            .as_ref()
+            .map(|uri| uri.borrow())
     }
 
-    ///  The location in the instance
-    pub const fn instance_location(&self) -> &JSONPointer {
+    /// The location in the instance
+    pub const fn instance_location(&self) -> &Location {
         &self.instance_location
     }
 }
@@ -308,16 +307,25 @@ impl<'a> From<&'a serde_json::Value> for Annotations<'a> {
     }
 }
 
-impl<'a> From<serde_json::Value> for Annotations<'a> {
+impl From<serde_json::Value> for Annotations<'_> {
     fn from(v: serde_json::Value) -> Self {
         Annotations(AnnotationsInner::Value(Box::new(v)))
     }
 }
 
-/// An error associated with an `OutputUnit`
+/// An error associated with an [`OutputUnit`]
 #[derive(serde::Serialize, Debug, Clone, PartialEq, Eq)]
 pub struct ErrorDescription(String);
 
+impl ErrorDescription {
+    /// Returns the inner [`String`] of the error description.
+    #[inline]
+    #[must_use]
+    pub fn into_inner(self) -> String {
+        self.0
+    }
+}
+
 impl fmt::Display for ErrorDescription {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.write_str(&self.0)
@@ -336,7 +344,7 @@ impl<'a> From<&'a str> for ErrorDescription {
     }
 }
 
-impl<'a> serde::Serialize for BasicOutput<'a> {
+impl serde::Serialize for BasicOutput<'_> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: serde::Serializer,
@@ -356,7 +364,7 @@ impl<'a> serde::Serialize for BasicOutput<'a> {
     }
 }
 
-impl<'a> serde::Serialize for AnnotationsInner<'a> {
+impl serde::Serialize for AnnotationsInner<'_> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: serde::Serializer,
@@ -369,14 +377,14 @@ impl<'a> serde::Serialize for AnnotationsInner<'a> {
     }
 }
 
-impl<'a> serde::Serialize for OutputUnit<Annotations<'a>> {
+impl serde::Serialize for OutputUnit<Annotations<'_>> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: serde::Serializer,
     {
         let mut map_ser = serializer.serialize_map(Some(4))?;
-        map_ser.serialize_entry("keywordLocation", &self.keyword_location)?;
-        map_ser.serialize_entry("instanceLocation", &self.instance_location)?;
+        map_ser.serialize_entry("keywordLocation", self.keyword_location.as_str())?;
+        map_ser.serialize_entry("instanceLocation", self.instance_location.as_str())?;
         if let Some(absolute) = &self.absolute_keyword_location {
             map_ser.serialize_entry("absoluteKeywordLocation", &absolute)?;
         }
@@ -391,8 +399,8 @@ impl serde::Serialize for OutputUnit<ErrorDescription> {
         S: serde::Serializer,
     {
         let mut map_ser = serializer.serialize_map(Some(4))?;
-        map_ser.serialize_entry("keywordLocation", &self.keyword_location)?;
-        map_ser.serialize_entry("instanceLocation", &self.instance_location)?;
+        map_ser.serialize_entry("keywordLocation", self.keyword_location.as_str())?;
+        map_ser.serialize_entry("instanceLocation", self.instance_location.as_str())?;
         if let Some(absolute) = &self.absolute_keyword_location {
             map_ser.serialize_entry("absoluteKeywordLocation", &absolute)?;
         }
